// =============================================================== //
//                                                                 //
//   File      : arb_misc.cxx                                      //
//   Purpose   : misc that doesnt fit elsewhere                    //
//                                                                 //
//   Coded by Ralf Westram (coder@reallysoft.de) in October 2012   //
//   Institute of Microbiology (Technical University Munich)       //
//   http://www.arb-home.de/                                       //
//                                                                 //
// =============================================================== //

#include "arb_misc.h"
#include "arb_msg.h"
#include "arb_file.h"
#include "arb_string.h"

#include <cmath>

// AISC_MKPT_PROMOTE:#ifndef _GLIBCXX_CSTDLIB
// AISC_MKPT_PROMOTE:#include <cstdlib>
// AISC_MKPT_PROMOTE:#endif

const char *GBS_readable_size(unsigned long long size, const char *unit_suffix) {
    // return human readable size information
    // returned string is maximal 6+strlen(unit_suffix) characters long
    // (using "b" as 'unit_suffix' produces '### b', '### Mb' etc)

    if (size<1000) return GBS_global_string("%llu %s", size, unit_suffix);

    const char *units = "kMGTPEZY"; // kilo, Mega, Giga, Tera, ... should be enough forever
    int i;

    for (i = 0; units[i]; ++i) {
        char unit = units[i];
        if (size<1000*1024) {
            double amount = size/(double)1024;
            if (amount<10.0)  return GBS_global_string("%4.2f %c%s", amount+0.005, unit, unit_suffix);
            if (amount<100.0) return GBS_global_string("%4.1f %c%s", amount+0.05, unit, unit_suffix);
            return GBS_global_string("%i %c%s", (int)(amount+0.5), unit, unit_suffix);
        }
        size /= 1024; // next unit
    }
    return GBS_global_string("MUCH %s", unit_suffix);
}

const char *GBS_readable_timediff(size_t seconds) {
    size_t mins  = seconds/60; seconds -= mins  * 60;
    size_t hours = mins/60;    mins    -= hours * 60;
    size_t days  = hours/24;   hours   -= days  * 24;

    const int   MAXPRINT = 40;
    int         printed  = 0;
    static char buffer[MAXPRINT+1];

    if (days>0)               printed += sprintf(buffer+printed, "%zud", days);
    if (printed || hours>0)   printed += sprintf(buffer+printed, "%zuh", hours);
    if (printed || mins>0)    printed += sprintf(buffer+printed, "%zum", mins);

    printed += sprintf(buffer+printed, "%zus", seconds);

    arb_assert(printed>0 && printed<MAXPRINT);

    return buffer;
}

const char *ARB_float_2_ascii(const float f) {
    /*! calculate the "best" ascii representation for float 'f'
     *  - smaller conversion error is better
     *  - shorter representation is better (for equal conversion errors)
     */

    const int   MAXSIZE = 50;
    static char result[MAXSIZE];
    char        buffer[MAXSIZE];

    int   printed_e = snprintf(result, MAXSIZE, "%e", f); arb_assert(printed_e<MAXSIZE);
    float back_e    = strtof(result, NULp);
    float diff_e    = fabsf(f-back_e);

    int   printed_g = snprintf(buffer, MAXSIZE, "%g", f); arb_assert(printed_g<MAXSIZE);
    float back_g    = strtof(buffer, NULp);
    float diff_g    = fabsf(f-back_g);

    if (diff_g<diff_e || (diff_g == diff_e && printed_g<printed_e)) {
        printed_e = printed_g;
        back_e    = back_g;
        diff_e    = diff_g;
        memcpy(result, buffer, printed_g+1);
    }

    int   printed_f = snprintf(buffer, MAXSIZE, "%f", f); arb_assert(printed_f<MAXSIZE);
    float back_f    = strtof(buffer, NULp);
    float diff_f    = fabsf(f-back_f);

    if (diff_f<diff_e || (diff_f == diff_e && printed_f<printed_e)) {
        memcpy(result, buffer, printed_f+1);
    }

    return result;
}

const char *ARB_getenv_ignore_empty(const char *envvar) {
    const char *result = getenv(envvar);
    return (result && result[0]) ? result : NULp;
}

char *ARB_executable(const char *exe_name, const char *path) {
    char       *buffer = ARB_alloc<char>(strlen(path)+1+strlen(exe_name)+1);
    const char *start  = path;
    int         found  = 0;

    while (!found && start) {
        const char *colon = strchr(start, ':');
        int         len   = colon ? (colon-start) : (int)strlen(start);

        memcpy(buffer, start, len);
        buffer[len] = '/';
        strcpy(buffer+len+1, exe_name);

        found = GB_is_executablefile(buffer);
        start = colon ? colon+1 : NULp;
    }

    char *executable = found ? ARB_strdup(buffer) : NULp;
    free(buffer);
    return executable;
}

// --------------------------------------------------------------------------------

char ARB_path_contains_unwanted_chars(const char *path) {
    if (strchr(path, ' ') != NULp) {
        return ' ';
    }
    return 0;
}

void ARB_warn_about_unwanted_chars(const char *path, const char *path_description) {
    // annoy user with warnings if paths contain unwanted characters.
    //
    // motivation: I tried to fix some scripts to correctly handle the case
    // where arb is installed in a path containing spaces.
    //
    // Since that is a bottomless pit, I decided to deny spaces there.

    char unwantedChar = ARB_path_contains_unwanted_chars(path);
    if (unwantedChar) {
        GB_warningf(
            "arb may not work as expected, because\n"
            "%s\n"
            "  (='%s')\n"
            "contains an unwanted character ('%c').",
            path_description,
            path,
            unwantedChar);
    }
}

// --------------------------------------------------------------------------------

#ifdef UNIT_TESTS
#ifndef TEST_UNIT_H
#include <test_unit.h>
#endif

#if 0
// simple test
#define TEST_EXPECT_FLOAT_2_ASCII(f,a) TEST_EXPECT_EQUAL(ARB_float_2_ascii(f), a)
#else
// also test back-conversion (ascii->float->ascii) is stable
#define TEST_EXPECT_FLOAT_2_ASCII(f,a) do{                          \
        TEST_EXPECT_EQUAL(ARB_float_2_ascii(f), a);                 \
        TEST_EXPECT_EQUAL(ARB_float_2_ascii(strtof(a, NULp)), a);   \
    }while(0)
#endif

__ATTR__REDUCED_OPTIMIZE void TEST_float_2_ascii() {
    TEST_EXPECT_FLOAT_2_ASCII(3.141592e+00, "3.141592");
    TEST_EXPECT_FLOAT_2_ASCII(3.141592,     "3.141592");
    TEST_EXPECT_FLOAT_2_ASCII(3.14159,      "3.14159");

    TEST_EXPECT_FLOAT_2_ASCII(0.1,           "0.1");
    TEST_EXPECT_FLOAT_2_ASCII(0.01,          "0.01");
    TEST_EXPECT_FLOAT_2_ASCII(0.001,         "0.001");
    TEST_EXPECT_FLOAT_2_ASCII(0.0001,        "0.0001");
    TEST_EXPECT_FLOAT_2_ASCII(0.00001,       "1e-05");
    TEST_EXPECT_FLOAT_2_ASCII(0.000001,      "1e-06");
    TEST_EXPECT_FLOAT_2_ASCII(0.0000001,     "1e-07");
    TEST_EXPECT_FLOAT_2_ASCII(0.00000001,    "1e-08");
    TEST_EXPECT_FLOAT_2_ASCII(0.000000001,   "1e-09");
    TEST_EXPECT_FLOAT_2_ASCII(0.0000000001,  "1e-10");
    TEST_EXPECT_FLOAT_2_ASCII(0.00000000001, "1e-11");

    TEST_EXPECT_FLOAT_2_ASCII(10,         "10");
    TEST_EXPECT_FLOAT_2_ASCII(100,        "100");
    TEST_EXPECT_FLOAT_2_ASCII(1000,       "1000");
    TEST_EXPECT_FLOAT_2_ASCII(10000,      "10000");
    TEST_EXPECT_FLOAT_2_ASCII(100000,     "100000");
    TEST_EXPECT_FLOAT_2_ASCII(1000000,    "1e+06");
    TEST_EXPECT_FLOAT_2_ASCII(10000000,   "1e+07");
    TEST_EXPECT_FLOAT_2_ASCII(100000000,  "1e+08");
    TEST_EXPECT_FLOAT_2_ASCII(1000000000, "1e+09");

    TEST_EXPECT_FLOAT_2_ASCII(3141592,      "3.141592e+06");
    TEST_EXPECT_FLOAT_2_ASCII(314159.2,     "3.141592e+05");
    TEST_EXPECT_FLOAT_2_ASCII(31415.92,     "3.141592e+04");
    TEST_EXPECT_FLOAT_2_ASCII(3141.592,     "3141.592041");
    TEST_EXPECT_FLOAT_2_ASCII(3141.592041,  "3141.592041");
    TEST_EXPECT_FLOAT_2_ASCII(314.1592,     "314.159210");
    TEST_EXPECT_FLOAT_2_ASCII(314.159210,   "314.159210");
    TEST_EXPECT_FLOAT_2_ASCII(31.41592,     "31.415920");
    TEST_EXPECT_FLOAT_2_ASCII(3.141592,     "3.141592");
    TEST_EXPECT_FLOAT_2_ASCII(.3141592,     "3.141592e-01");
    TEST_EXPECT_FLOAT_2_ASCII(.03141592,    "3.141592e-02");
    TEST_EXPECT_FLOAT_2_ASCII(.003141592,   "3.141592e-03");
    TEST_EXPECT_FLOAT_2_ASCII(.0003141592,  "3.141592e-04");
    TEST_EXPECT_FLOAT_2_ASCII(.00003141592, "3.141592e-05");
    TEST_EXPECT_FLOAT_2_ASCII(M_PI,         "3.141593");

    TEST_EXPECT_FLOAT_2_ASCII(1/2.0, "0.5");
    TEST_EXPECT_FLOAT_2_ASCII(1/3.0, "3.333333e-01");
    TEST_EXPECT_FLOAT_2_ASCII(1/4.0, "0.25");
    TEST_EXPECT_FLOAT_2_ASCII(1/5.0, "0.2");
    TEST_EXPECT_FLOAT_2_ASCII(1/6.0, "1.666667e-01");

    TEST_EXPECT_FLOAT_2_ASCII(37550000.0,  "3.755e+07");
    TEST_EXPECT_FLOAT_2_ASCII(3755000.0,   "3.755e+06");
    TEST_EXPECT_FLOAT_2_ASCII(375500.0,    "375500");
    TEST_EXPECT_FLOAT_2_ASCII(37550.0,     "37550");
    TEST_EXPECT_FLOAT_2_ASCII(3755.0,      "3755");
    TEST_EXPECT_FLOAT_2_ASCII(375.5,       "375.5");
    TEST_EXPECT_FLOAT_2_ASCII(37.55,       "37.55");
    TEST_EXPECT_FLOAT_2_ASCII(3.755,       "3.755");
    TEST_EXPECT_FLOAT_2_ASCII(0.3755,      "0.3755");
    TEST_EXPECT_FLOAT_2_ASCII(0.03755,     "0.03755");
    TEST_EXPECT_FLOAT_2_ASCII(0.003755,    "0.003755");
    TEST_EXPECT_FLOAT_2_ASCII(0.0003755,   "0.0003755");
    TEST_EXPECT_FLOAT_2_ASCII(0.00003755,  "3.755e-05");
    TEST_EXPECT_FLOAT_2_ASCII(0.000003755, "3.755e-06");

    TEST_EXPECT_FLOAT_2_ASCII(1000.0*1000.0*1000.0,    "1e+09");
    TEST_EXPECT_FLOAT_2_ASCII(25000.0*25000.0*25000.0, "1.5625e+13");
}

// ------------------------------------------------------------
// test to ensure sanitizers work as expected

#if 0
void TEST_fail_address_sanitizer() {
    static int array[5];
    array[2] = 1;
    array[5] = 1; // <- fails with AddressSanitizer

    printf("array[5]=%i\n", array[5]);
}
#endif

#if 0
void TEST_fail_undef_sanitizer() {
    // error below are not reported if AddressSanitizer bails out (TEST_fail_address_sanitizer)
    int x  = 7;
    int y1 = -1;

    int s = x<<y1; // runtime error with ubsan: shift exponent -1 is negative (does not terminate)
    printf("s=%i\n", s);

    int o = INT_MAX;
    int u = INT_MIN;
    o++; // runtime error: signed integer overflow
    u--; // runtime error: signed integer overflow
    printf("o=%i u=%i\n", o, u);

#if 0
    int y2 = 0;
    int z1 = x/y1;
    int z2 = x/y2; // runtime error with ubsan: division by zero (terminates with SEGV; also w/o sanitizers)
    printf("z1=%i z2=%i\n", z1, z2);
#endif
}
#endif

#if 0
void TEST_fail_leak_sanitizer() {
    int *p = new int[5]; // <- fails with LeakSanitizer (only reported if AddressSanitizer does not bail out (TEST_fail_address_sanitizer))
    printf("p[3]=%i\n", p[3]);
}
#endif

// ------------------------------------------------------------

#include "StrUniquifier.h"

void TEST_StrUniquifier() {
    StrUniquifier uniq("->");
    TEST_EXPECT_EQUAL(uniq.make_unique_key("hey"), "hey");
    TEST_EXPECT_EQUAL(uniq.make_unique_key("hey"), "hey->2");
    TEST_EXPECT_EQUAL(uniq.make_unique_key("Hey"), "Hey");
    TEST_EXPECT_EQUAL(uniq.make_unique_key("Hey"), "Hey->2");

    TEST_EXPECT_EQUAL(uniq.make_unique_key(""), "");
    TEST_EXPECT_EQUAL(uniq.make_unique_key(""), "->2");

    StrUniquifier fresh(".");
    TEST_EXPECT_EQUAL(fresh.make_unique_key(""), "");
    TEST_EXPECT_EQUAL(fresh.make_unique_key(""), ".2");
}

// ------------------------------------------------------------

#include <arb_error.h>
#include <ErrorOrType.h>

static const char *SOME_ERROR = "whatever";

static void drop_unused_ARB_ERROR() {
    ARB_ERROR e0;
}
static void drop_ARB_ERROR() {
    ARB_ERROR e1 = SOME_ERROR;
}
static void overwrite_ARB_ERROR() {
    ARB_ERROR e1 = SOME_ERROR;
    ARB_ERROR e2 = SOME_ERROR;

    e1 = e2;           // expect this command to trigger assertion
    ARB_SIGSEGV(true); // -> this is never reached
}

static ARB_ERROR forwardError(ARB_ERROR err) {
    return err;
}
static ARB_ERROR annotateError(ARB_ERROR err) {
    if (err) {
        ARB_ERROR ann_err = GBS_global_string("annotated '%s'", err.deliver());
        err               = ann_err;
    }
    return err;
}

void TEST_misuse_ARB_ERROR_crashtest() {
    TEST_EXPECT_CODE_ASSERTION_FAILS(drop_unused_ARB_ERROR);
    TEST_EXPECT_CODE_ASSERTION_FAILS(drop_ARB_ERROR);
    TEST_EXPECT_CODE_ASSERTION_FAILS(overwrite_ARB_ERROR);
}
void TEST_ARB_ERROR() {
    // unused error:
    {
        ARB_ERROR e0;
        TEST_EXPECT_NULL(e0.deliver());
    }

    // simple error delivery:
    {
        ARB_ERROR e1 = SOME_ERROR;
        TEST_EXPECT_EQUAL(e1.deliver(), SOME_ERROR);
    }

    // deliver reassigned error (copy ctor):
    {
        ARB_ERROR e1 = SOME_ERROR;
        ARB_ERROR e2(e1);
        TEST_EXPECT_EQUAL(e2.deliver(), SOME_ERROR);
    }
    // deliver reassigned error (also copy ctor):
    {
        ARB_ERROR e1 = SOME_ERROR;
        ARB_ERROR e2 = e1; // assign in definition == copy-ctor
        TEST_EXPECT_EQUAL(e2.deliver(), SOME_ERROR);
    }
    // deliver reassigned error (op=):
    {
        ARB_ERROR e1 = SOME_ERROR;
        ARB_ERROR e2;
        e2 = e1; // real assignment
        TEST_EXPECT_EQUAL(e2.deliver(), SOME_ERROR);
    }

    // deliver error forwarded through function:
    {
        ARB_ERROR e0;
        ARB_ERROR e1 = SOME_ERROR;
        ARB_ERROR f0 = forwardError(e0);
        ARB_ERROR f1 = forwardError(e1);
        TEST_EXPECT_NULL(f0.deliver());
        TEST_EXPECT_EQUAL(f1.deliver(), SOME_ERROR);
    }
    // forward and annotate error:
    {
        ARB_ERROR e0;
        ARB_ERROR e1 = SOME_ERROR;
        ARB_ERROR a0 = annotateError(e0);
        ARB_ERROR a1 = annotateError(e1);
        TEST_EXPECT_NULL(a0.deliver());
        TEST_EXPECT_EQUAL(a1.deliver(), "annotated 'whatever'");
    }
}

typedef ErrorOr<int> ErrorOrInt;

static void drop_ErrorOr0() {
    ErrorOrInt e(NULp, 42);
}
static void drop_ErrorOr1() {
    ErrorOrInt e(SOME_ERROR, 0);
}
static void retrieve_unchecked_ErrorOr0() {
    ErrorOrInt e(NULp, 21);
    int v = e.getValue(); // getValue w/o previous mandatory hasError
    TEST_EXPECT_EQUAL(v, 21);
}
static void retrieve_unchecked_ErrorOr1() {
    ErrorOrInt e(SOME_ERROR, 0);
    ARB_ERROR  ae = e.getError(); // getError w/o previous mandatory hasError
    TEST_REJECT_NULL(ae.deliver());
}

static ErrorOrInt generateErrorOr(int i) {
    return ErrorOrInt(NULp, i);
}
static ErrorOrInt generateErrorOr(const char *msg) {
    return ErrorOrInt(msg, -1);
}

void TEST_misuse_ErrorOr_crashtest() {
    TEST_EXPECT_CODE_ASSERTION_FAILS(drop_ErrorOr0);
    TEST_EXPECT_CODE_ASSERTION_FAILS(drop_ErrorOr1);
    TEST_EXPECT_CODE_ASSERTION_FAILS(retrieve_unchecked_ErrorOr0);
    TEST_EXPECT_CODE_ASSERTION_FAILS(retrieve_unchecked_ErrorOr1);
}
void TEST_ErrorOr() {
    // simple uses (create, check, use + destroy):
    {
        ErrorOrInt e(SOME_ERROR, 0);
        TEST_EXPECT(e.hasError());
        TEST_EXPECT_EQUAL(e.getError().deliver(), SOME_ERROR);
    }
    {
        ErrorOrInt v(NULp, 7);
        TEST_EXPECT(v.hasValue());
        TEST_EXPECT_EQUAL(v.getValue(), 7);
    }

    // test copy-construction:
    {
        ErrorOrInt e(SOME_ERROR, 0);
        ErrorOrInt c = e;
        TEST_EXPECT(c.hasError());
        TEST_EXPECT_EQUAL(c.getError().deliver(), SOME_ERROR);
    }
    {
        ErrorOrInt v(NULp, 17);
        ErrorOrInt c = v;
        TEST_EXPECT(c.hasValue());
        TEST_EXPECT_EQUAL(c.getValue(), 17);
    }
    {
        ErrorOrInt v = generateErrorOr(17);
        TEST_EXPECT(v.hasValue());
        TEST_EXPECT_EQUAL(v.getValue(), 17);
    }
    {
        ErrorOrInt e = generateErrorOr(SOME_ERROR);
        TEST_EXPECT(e.hasError());
        TEST_EXPECT_EQUAL(e.getError().deliver(), SOME_ERROR);
    }
}

#endif // UNIT_TESTS

// --------------------------------------------------------------------------------

